#ifndef INCLUDE_UTILS_SHARD_H_
#define INCLUDE_UTILS_SHARD_H_

#include <google/protobuf/message.h>
#include <fstream>
#include <string>
#include <unordered_set>


using google::protobuf::Message;

namespace singa {

/**
 * Data shard stores training/validation/test tuples.
 * Every worker node should have a training shard (validation/test shard
 * is optional). The shard file for training is
 * singa::Cluster::workspace()/train/shard.dat; The shard file for validation
 * is singa::Cluster::workspace()/train/shard.dat; Similar path for test.
 *
 * shard.dat consists of a set of unordered tuples. Each tuple is
 * encoded as [key_len key record_len val] (key_len and record_len are of type
 * uint32, which indicate the bytes of key and record respectively.
 *
 * When Shard obj is created, it will remove the last key if the record size and
 * key size do not match because the last write of tuple crashed.
 *
 * TODO
 * 1. split one shard into multile shards.
 * 2. add threading to prefetch and parse records
 *
 */
class DataShard {
 public:
  enum {
    //!< read only mode used in training
    kRead=0,
    //!< write mode used in creating shard (will overwrite previous one)
    kCreate=1,
    //!< append mode, e.g. used when previous creating crashes
    kAppend=2
  };

 public:
  /**
   * Init the shard obj.
   * @folder shard folder (path excluding shard.dat) on worker node
   * @mode shard open mode, Shard::kRead, Shard::kWrite or Shard::kAppend
   * @bufsize batch bufsize bytes data for every disk op (read or write),
   * default is 100MB
   */
  DataShard(std::string folder, char mode, int capacity=104857600);
  ~DataShard();

  /**
   * read next tuple from the shard.
   * @key key
   * @param val record of type Message
   * @return true if read success otherwise false, e.g., the tuple was not
   * inserted completely.
   */
  bool Next(std::string *key, Message* val);
  /**
   * read next tuple from the shard.
   * @key key tuple key
   * @param val record of type string
   * @return true if read success otherwise false, e.g., the tuple was not
   * inserted completely.
   */
  bool Next(std::string *key, std::string* val);

  /**
   * Append one tuple to the shard.
   * @param key e.g., image path
   * @param val
   * @return reture if sucess, otherwise false, e.g., inserted before
   */
  bool Insert(const std::string& key, const Message& tuple);
  /**
   * Append one tuple to the shard.
   * @param key e.g., image path
   * @param val
   * @return reture if sucess, otherwise false, e.g., inserted before
   */
  bool Insert(const std::string& key, const std::string& tuple);
  /**
   * Move the read pointer to the head of the shard file.
   * Used for repeated reading.
   */
  void SeekToFirst();
  /**
   * Flush buffered data to disk.
   * Used only for kCreate or kAppend.
   */
  void Flush() ;
  /**
   * Iterate through all tuples to get the num of all tuples.
   * @return num of tuples
   */
  const int Count();
  /**
   * @return path to shard file
   */
  const std::string path(){
    return path_;
  }

 protected:
  /**
   * Read the next key and prepare buffer for reading value.
   * @param key
   * @return length (i.e., bytes) of value field.
   */
  int Next(std::string *key);
  /**
   * Setup the disk pointer to the right position for append in case that
   * the pervious write crashes.
   * @param path shard path.
   * @return offset (end pos) of the last success written record.
   */
  int PrepareForAppend(std::string path);
  /**
   * Read data from disk if the current data in the buffer is not a full field.
   * @param size size of the next field.
   */
  bool PrepareNextField(int size);

 private:
  char mode_;
  std::string path_;
  // either ifstream or ofstream
  std::fstream fdat_;
  // to avoid replicated record
  std::unordered_set<std::string> keys_;
  // internal buffer
  char* buf_;
  // offset inside the buf_
  int offset_;
  // allocated bytes for the buf_
  int capacity_;
  // bytes in buf_, used in reading
  int bufsize_;
};
} /* singa */
#endif  // INCLUDE_UTILS_SHARD_H_
